cimport numpy as np
import numpy as np




cdef unsigned int hamming_distance(str seq1, str seq2, unsigned int limit=0):
    """
    Returns the Hamming distance between equal-length sequences.
    :param seq1: first sequence.
    :param seq2: second sequence.
    :param limit: max distance limit before aborting (returning limit + 1).
    :return: the edit distance.
    """
    cdef unsigned int i
    cdef unsigned int sum = 0
    for i in xrange(len(seq1)):
        if seq1[i] != seq2[i]:
            sum += 1
        if limit > 0 and sum > limit:
            return limit + 1
    return sum



cdef unsigned int levenshtein_distance(str seq1, str seq2, unsigned int limit=0):
    """
    Returns the Levenshtein distance between two sequences. Lengths do no need to be equal, as indels are allowed.
    :param seq1: first sequence.
    :param seq2: second sequence.
    :param limit: max distance limit before aborting (returning limit + 1).
    :return: the edit distance.
    """
    cdef list one_ago = None
    cdef list this_row = range(1, len(seq2) + 1) + [0]
    cdef int x
    cdef int y
    cdef int del_cost
    cdef int add_cost
    cdef int sub_cost
    for x in xrange(len(seq1)):
        one_ago = this_row
        this_row = [0] * len(seq2) + [x + 1]
        for y in xrange(len(seq2)):
            del_cost = one_ago[y] + 1
            add_cost = this_row[y - 1] + 1
            sub_cost = one_ago[y - 1] + (seq1[x] != seq2[y])
            this_row[y] = min(del_cost, add_cost, sub_cost)
        if limit > 0 and x > limit and min(this_row) > limit:
            return limit + 1
    return this_row[len(seq2) - 1]



cdef list subglobal_distance(str s1, str s2):
    """
    Computes the edit distance for a sub-global (local?) alignment
    of a sequence s2 against a sequence s1.
    Mismatches and indels both score as 1. Overhanging parts of s1 do not count.
    :param s1: the longer (probe) sequence.
    :param s2: the shorter sought sequence.
    :return: [edit distance, last probe pos, s1 insertions, s2 insertions], where indels are estimates from a
    a minimal alignment, but indel counts are not necessarily minimal.
    """

    cdef unsigned int xLen = len(s1)
    cdef unsigned int yLen = len(s2)
    if xLen < yLen:
        raise ValueError("Sub-global edit distance is undefined for sequences where the probe is shorter than the aligned sequence.")

    #cdef list d = list()
    #d.append(list())
    #for y in xrange(0, yLen+1):
    #    d[0].append(y)  # To ensure all of s2 is spanned.

    # Perform DP.
    #for x in xrange(1, xLen+1):
    #    d.append(list())
    #    d[x].append(0)
    #    # Fill matrix.
    #    for y in xrange(1, yLen+1):
    #        d[x].append(min( min(d[x-1][y]+1, d[x][y-1]+1), d[x-1][y-1] + int(s1[x-1] != s2[y-1]) ))


    cdef unsigned int x
    cdef unsigned int y

    cdef np.ndarray[np.uint32_t, ndim=2] d = np.empty([xLen+1, yLen+1], dtype=np.uint32)
    for x in xrange(0, xLen+1):
        d[x,0] = 0
    for y in xrange(1, yLen+1):
        d[0,y] = y # To ensure all of s2 is spanned.

    # Perform DP.
    for x in xrange(1, xLen+1):
        # Fill matrix.
        for y in xrange(1, yLen+1):
            d[x,y] = min( min(d[x-1,y]+1, d[x,y-1]+1), d[x-1,y-1] + int(s1[x-1] != s2[y-1]) )


    #for x in xrange(xLen + 1):
    #    for y in xrange(yLen + 1):
    #        print d[x,y],
    #        print " ",
    #    print


    # Find min for sub-global alignment so that all of s2 is covered, but not necessarily all of s1 sequence.
    cdef unsigned int mini = 100000
    cdef unsigned int iPos = 0
    cdef unsigned int i = xLen
    cdef unsigned int j
    while i > 0:
        if d[i,yLen] < mini:
            mini = d[i,yLen]
            iPos = i
        i -= 1

    # Compute backtracking for indels.
    cdef unsigned int s2ins = 0
    cdef unsigned int s1ins = 0

    # Last elements of alignment.
    i = iPos
    j = yLen

    #print str(iPos) + " " + str(mini)

    # Backtrack.
    while (i > 0) and (j > 0):
        if d[i,j] == d[i-1,j-1] + int(s1[i-1] != s2[j-1]):
            # Subst.
            i -= 1
            j -= 1
        elif d[i,j] == d[i-1,j] + 1:
            # Add probe insertion. NOTE: Heavier weighting since this elif appears before below elif.
            s1ins += 1
            i -= 1
        else:
            # Add sought insertion.
            s2ins += 1
            j -= 1

    return [mini, iPos-1, s1ins, s2ins]   # Indexed from 1...



cimport taggd.misc.kmer_utils as ku
cimport taggd.misc.distance_metrics as dm
from cpython cimport bool


# User options
cdef dict true_barcodes
cdef unsigned int k
cdef unsigned int max_edit_distance
cdef str metric
cdef unsigned int slider_increment
cdef unsigned int pre_overhang
cdef unsigned int post_overhang


# Dictionary
cdef dict kmer2seq = None


# Metrics
cdef int SUBGLOBAL
cdef int LEVENSHTEIN
cdef int HAMMING
cdef int metric_choice



def init(\
        dict true_barcodes_, \
        unsigned int k_,
        unsigned int max_edit_distance_, \
        str metric_, \
        unsigned int slider_increment_, \
        unsigned int pre_overhang_, \
        unsigned int post_overhang_\
    ):
    """
    Initializes settings (global variables).
    :param true_barcodes_: true barcodes dict to attributes.
    :param k_: the k-mer length used in the search algorithm.
    :param max_edit_distance_: max allowed distance.
    :param metric_: distance metric to use, Subglobal, Levenshtein or Hamming.
    :param slider_increment_: the number of increments for kmer search. The default, 0 sets it to k.
    :param pre_overhang: pre-read flanking bases.
    :param post_overhang: post-read flanking bases.
    """

    global true_barcodes
    true_barcodes = true_barcodes_
    global k
    k = k_
    global max_edit_distance
    max_edit_distance = max_edit_distance_
    global metric
    metric = metric_
    global slider_increment
    slider_increment = slider_increment_
    global pre_overhang
    pre_overhang = pre_overhang_
    global post_overhang
    post_overhang = post_overhang_


    # Create k-mer mappings with ALL kmers
    global kmer2seq
    seq2kmer, kmer2seq = ku.get_kmers_dicts(true_barcodes.keys(), k, False, 1)

    # Metrics
    global SUBGLOBAL
    global LEVENSHTEIN
    global HAMMING
    SUBGLOBAL, LEVENSHTEIN, HAMMING = range(0, 3)
    global metric_choice
    if metric == "Subglobal":
        metric_choice = SUBGLOBAL
    elif metric == "Levenshtein":
        metric_choice = LEVENSHTEIN
    elif metric == "Hamming":
        metric_choice = HAMMING
    else:
        raise ValueError("Invalid distance metric specified")




cdef dict get_candidates(str read_barcode):
    """
    Returns candidate barcodes for a read barcode.
    :param read_barcode: read barcode.
    :return: the candidates as a dict with <barcode, read kmer hits>.
    """
    cdef dict candidates = dict()
    cdef list kmers_offsets = ku.get_kmers(read_barcode, k, False, slider_increment)
    cdef str kmer
    cdef unsigned int offset
    cdef dict hits
    cdef str hit
    cdef list hit_offsets
    cdef unsigned int hit_offset
    cdef unsigned int penalty = 0
    cdef unsigned int min_penalty = 0
    # For each kmer in read (typically incremented by k positions at a time).
    for kmer, offset in kmers_offsets:
        if kmer in kmer2seq:
            hits = kmer2seq[kmer]
        else:
            continue
        # For each true barcode containing read's kmer.
        for hit, hit_offsets in hits.iteritems():
            min_penalty = 100000000
            # For each position that kmer occurred in the true barcode.
            for hit_offset in hit_offsets:
                # Kmer may be shifted overhang positions without penalty, due to subglobal alignment.
                penalty = max(0, abs(int(offset) - int(hit_offset)) - int(pre_overhang) - int(post_overhang))
                if penalty < min_penalty:
                    min_penalty = penalty
            if hit in candidates:
                candidates[hit] += min_penalty
            else:
                candidates[hit] = min_penalty
    # Clear out all candidates with a forced offset penalty greater than the max edit distance:
    for hit in candidates.keys():
        if candidates[hit] > max_edit_distance:
            del candidates[hit]
    return candidates




cdef list get_distances(str read_barcode, dict candidates):
    """
    Returns all qualified hits ordered by distance.
    :param read_barcode: the read barcode.
    :param candidates: the candidate dictionary.
    :return: a list of tuples, (barcode,distance,lastpos,ins,del).
    """
    cdef list qual_hits = []
    cdef str candidate = None
    cdef unsigned int penalty = 0
    cdef unsigned int dist = 0
    cdef unsigned int max_lim = 0
    cdef unsigned int read_last_pos = 0
    cdef unsigned int a = 0
    cdef unsigned int b = 0

    for candidate, penalty in candidates.iteritems():
        if metric_choice == SUBGLOBAL:
            dist, read_last_pos, a, b = dm.subglobal_distance(read_barcode, candidate)
            if dist <= max_edit_distance:
                qual_hits.append((candidate, dist, str(read_last_pos),  str(a), str(b)))
        elif metric_choice == LEVENSHTEIN:
            # Account for the added overhang!
            # Note: This does NOT equate subglobal and may miss cases subglobal would catch!
            max_lim = max_edit_distance + max(0, len(read_barcode) - len(candidate))
            dist = dm.levenshtein_distance(read_barcode, candidate, max_lim)
            if dist <= max_lim:
                qual_hits.append((candidate, min(dist, max_edit_distance), "-", "-", "-"))
        elif metric_choice == HAMMING:
            dist = dm.hamming_distance(read_barcode, candidate, max_edit_distance)
            if dist <= max_edit_distance:
                qual_hits.append((candidate, dist, "-", "-", "-"))
        else:
            raise ValueError("Invalid distance metric specified")
    return qual_hits




cdef list get_top_hits(list qual_hits):
    """
    Returns the top hits.
    :param qual_hits: the list of qualified hits.
    :return: the top hits.
    """
    if len(qual_hits) == 0:
        return None
    # Find smallest.
    cdef unsigned int mini = 10000000
    cdef str cand = None
    cdef unsigned int dist = 0
    cdef str a = None
    cdef str b = None
    for cand, dist, last_pos, a, b in qual_hits:
        if dist < mini:
            mini = dist
    # Filter out the smallest
    cdef list top_hits = []
    for cand, dist, last_pos, a, b in qual_hits:
        if dist == mini:
            top_hits.append((cand, dist, last_pos, a, b))
    return top_hits



"""
Main functions for carrying out the demultiplexing with
multithreading.
"""

import sys
import multiprocessing as mp
import random
import taggd.core.demultiplex_search_functions as srch
cimport taggd.core.demultiplex_search_functions as srch
import taggd.io.reads_reader_writer as rw
from taggd.misc.counter import *
from cpython cimport bool


# Global variables for user options
cdef dict true_barcodes
cdef str reads_infile
cdef str outfile_prefix
cdef unsigned int max_edit_distance
cdef unsigned int start_position
cdef unsigned int barcode_length
cdef unsigned int pre_overhang
cdef unsigned int post_overhang
cdef str seed
cdef bool no_multiprocessing
cdef bool only_output_matched


# Reader-writer
cdef object re_wr
cdef object f_match = None
cdef object f_res = None
cdef object f_ambig = None
cdef object f_unmatch = None


# Global variables for stats
cdef object stats_total_reads
cdef object stats_unmatched
cdef object stats_perfect_matches
cdef object stats_imperfect_unambiguous_matches
cdef object stats_imperfect_ambiguous_matches
cdef list stats_edit_distance_counts


# Global variables for multithreading.
cdef object manager
cdef object q
cdef object pool
cdef object watcher


# Global variables for match types.
cdef int KILL, UNMATCHED, MATCHED_PERFECTLY, MATCHED_UNAMBIGUOUSLY, MATCHED_AMBIGUOUSLY


def init(\
        dict true_barcodes_, \
        str reads_infile_, \
        str outfile_prefix_, \
        unsigned int max_edit_distance_, \
        unsigned int start_position_, \
        unsigned int pre_overhang_, \
        unsigned int post_overhang_, \
        str seed_, \
        bool no_multiprocessing_,\
        bool only_output_matched_\
    ):
    """
    Initializes settings (global variables).
    :param true_barcodes_: barcodes dict to attributes.
    :param reads_infile_: input file with read sequences (containing read barcodes).
    :param outfile_prefix_: output file prefix.
    :param max_edit_distance_: max allowed distance.
    :param start_position_: the start position.
    :param pre_overhang_: how many flanking bases pre for indels. Not applicable for Hamming distance.
    :param post_overhang_: how many flanking bases post for indels. Not applicable for Hamming distance.
    :param seed_: PRNG seed.
    :param no_multiprocessing: No multiprocessing.
    :param only_output_matched: Only matched file out.
    """

    global true_barcodes
    true_barcodes = true_barcodes_
    global reads_infile
    reads_infile = reads_infile_
    global outfile_prefix
    outfile_prefix = outfile_prefix_
    global max_edit_distance
    max_edit_distance = max_edit_distance_
    global start_position
    start_position = start_position_
    global pre_overhang
    pre_overhang = pre_overhang_
    global post_overhang
    post_overhang = post_overhang_
    global seed
    seed = seed_
    random.seed(seed)
    global no_multiprocessing
    no_multiprocessing = no_multiprocessing_
    global only_output_matched
    only_output_matched = only_output_matched_

    # Read chip file
    global barcode_length
    barcode_length = len(true_barcodes.keys()[0])

    # Reader writer
    global re_wr
    re_wr = rw.ReadsReaderWriter(reads_infile)

    # Stats
    global stats_total_reads
    stats_total_reads = Counter(0)
    global stats_unmatched
    stats_unmatched = Counter(0)
    global stats_perfect_matches
    stats_perfect_matches = Counter(0)
    global stats_imperfect_unambiguous_matches
    stats_imperfect_unambiguous_matches = Counter(0)
    global stats_imperfect_ambiguous_matches
    stats_imperfect_ambiguous_matches = Counter(0)
    global stats_edit_distance_counts
    stats_edit_distance_counts = []
    cdef unsigned int i
    for i in xrange(max_edit_distance+1):
        stats_edit_distance_counts.append(Counter(0))

    # Match types
    global KILL
    global UNMATCHED
    global MATCHED_PERFECTLY
    global MATCHED_UNAMBIGUOUSLY
    global MATCHED_AMBIGUOUSLY
    KILL, UNMATCHED, MATCHED_PERFECTLY, MATCHED_UNAMBIGUOUSLY, MATCHED_AMBIGUOUSLY = range(-1,4)




def demultiplex():
    """Demultiplexes the contents of a reads file."""

    # Open files.
    global f_match
    f_match = re_wr.get_writer(outfile_prefix + ".matched." + re_wr.get_format())
    if not only_output_matched:
        global f_res
        f_res = open(outfile_prefix + ".results.tsv", 'w')
        f_res.write("#Annotation\tMatch_result\tBarcode\tEdit_distance\tAmbiguous_top_hits\tQualified_candidates\tRaw_candidates\tLast_position\tApprox_insertions\tApprox_deletions\n")
        global f_ambig
        f_ambig = re_wr.get_writer(outfile_prefix + ".ambiguous." + re_wr.get_format())
        global f_unmatch
        f_unmatch = re_wr.get_writer(outfile_prefix + ".unmatched." + re_wr.get_format())
    else:
        global f_res
        f_res = None
        global f_ambig
        f_ambig = None
        global f_unmatch
        f_unmatch = None

    # Demultiplex
    if no_multiprocessing:
        __demultiplex_linearly()
    else:
        __demultiplex_mp()



def __demultiplex_linearly():
    """
    Demultiplexes the contents of a reads file in a linear manner.
    """
    global manager
    manager = mp.Manager()
    global q
    q = manager.Queue()

    cdef object rec = None
    for rec in re_wr.reader_open():
        __demultiplex_record_wrapper(rec)
    re_wr.reader_close()
    q.put((None, KILL, None, None))
    __listener()



def __demultiplex_mp():
    """
    Demultiplexes the contents of a reads file in a multithreaded manner.
    """

    # Must use Manager queue here, or will not work
    global manager
    manager = mp.Manager()
    global q
    q = manager.Queue()
    global pool
    pool = mp.Pool(mp.cpu_count())

    # Put listener to work first
    global watcher
    watcher = pool.apply_async(__listener, ())

    # Fire off workers
    cdef list jobs = []
    cdef object job = None
    cdef object rec = None
    for rec in re_wr.reader_open():
        job = pool.apply_async(__demultiplex_record_wrapper, (rec,))
        jobs.append(job)
    re_wr.reader_close()

    # Collect results from the workers through the pool result queue
    for job in jobs:
        job.get()

    # Now we are done, kill the listener
    q.put((None, KILL, None, None))
    pool.close()



def __listener():
    """
    Opens files, listens for messages on the q, writes output to files.
    """

    cdef object rec = None
    cdef int result = -1
    cdef str bcseq = None
    cdef object bc = None
    cdef list props = []
    cdef list tags = []
    cdef unsigned int i = 0
    while True:

        # Extract record
        (rec, result, bcseq, props) = q.get()

        # End of line.
        if result == KILL:
            break

        # Write to info file.
        if not only_output_matched:
            f_res.write("%s\t%s\t%s\t%s\n" % (rec.annotation, __match_type_to_str(result), bcseq, "\t".join(props)))

        # No match.
        if result == UNMATCHED:
            if not only_output_matched:
                re_wr.write_record(f_unmatch, rec)
            continue

        # Append record with properties. B0:Z:Barcode, B1:Z:Prop1, B2:Z:prop3 ...
        bc = true_barcodes[bcseq]
        tags = list()
        tags.append(("B0:Z", bcseq))
        for i in xrange(len(bc.attributes)):
            tags.append(("B" + str(i+1) + ":Z", bc.attributes[i]))
        rec.add_tags(tags)

        # Write to file.
        if result == MATCHED_PERFECTLY:
            re_wr.write_record(f_match, rec)
        elif result == MATCHED_UNAMBIGUOUSLY:
            re_wr.write_record(f_match, rec)
        elif result == MATCHED_AMBIGUOUSLY and not only_output_matched:
            re_wr.write_record(f_ambig, rec)
        else:
            continue

    # Close all files. DO HERE, WHEN WE'RE SURE MP IS FINISHED!
    f_match.close()
    if not only_output_matched:
        f_res.close()
        f_ambig.close()
        f_unmatch.close()



def __demultiplex_record_wrapper(object rec):
    """
    Wrapper for cdef:ed multithreading function.
    """
    return __demultiplex_record(rec)




cdef bool __demultiplex_record(object rec):
    """
    Demultiplexes a record.
    :param rec: the record.
    :return: true if mapped.
    """
    stats_total_reads.increment()

    cdef str read_barcode = rec.sequence[int(start_position):int(start_position+barcode_length)]
    cdef dict candidates = None
    cdef list qual_hits = None
    cdef list top_hits = None
    cdef str bcseq = None
    cdef unsigned int dist = 0
    cdef str last_pos = None
    cdef str a = None
    cdef str b = None
    if read_barcode in true_barcodes:
        # PERFECT MATCH
        # Record   Match type   Barcode   [Edit distance, Ambiguous top hits, Qualified candidates, Raw candidates, Last_pos, Insertions read, Insertions true]
        q.put((rec, MATCHED_PERFECTLY, read_barcode, ["0", "1", "1", "-", str(barcode_length-1), "0", "0"]))
        stats_perfect_matches.increment()
        stats_edit_distance_counts[0].increment()
        return True
    else:
        # Include overhang.
        read_barcode = rec.sequence[int(start_position-pre_overhang):min(len(rec.sequence), int(start_position+barcode_length+post_overhang))]

        # Narrow down hits.
        candidates = srch.get_candidates(read_barcode)
        qual_hits = srch.get_distances(read_barcode, candidates)
        top_hits = srch.get_top_hits(qual_hits)

        if not top_hits:
            # NO MATCH.
            # Record   Match type   Barcode   [Edit distance, Ambiguous top hits, Qualified candidates, Raw candidates, Last_pos, Insertions read, Insertions true]
            q.put((rec, UNMATCHED, "-", ["-", "0", str(len(qual_hits)), str(len(candidates)), "-", "-", "-"]))
            stats_unmatched.increment()
            return False
        else:
            if len(top_hits) == 1:
                # SINGLE MATCH.
                bcseq, dist, last_pos, a, b = top_hits[0]
                # Record   Match type   Barcode   [Edit distance, Ambiguous top hits, Qualified candidates, Raw candidates, Last_pos, Insertions read, Insertions true]
                q.put((rec, MATCHED_UNAMBIGUOUSLY, bcseq, [str(dist), str(len(top_hits)), str(len(qual_hits)), str(len(candidates)), last_pos, a, b]))
                stats_imperfect_unambiguous_matches.increment()
                stats_edit_distance_counts[dist].increment()
                return True
            else:
                # AMBIGUOUS MATCHES.
                # Multiple best hits. Shuffle results to avoid biases.
                random.shuffle(top_hits)
                for (bcseq, dist, last_pos, a, b) in top_hits:
                    # Record   Match type   Barcode   [Edit distance, Ambiguous top hits, Qualified candidates, Raw candidates, Last_pos, Insertions read, Insertions true]
                    q.put((rec, MATCHED_AMBIGUOUSLY, bcseq, [str(dist), str(len(top_hits)), str(len(qual_hits)), str(len(candidates)), last_pos, a, b]))
                stats_imperfect_ambiguous_matches.increment()
                stats_edit_distance_counts[dist].increment()
                return True



def print_pre_stats():
    """
    Prints pre stats
    """
    print "# Absolute output prefix path: " + outfile_prefix
    if only_output_matched:
        print "# Only writing matched reads."
    print "# Reads format: " + re_wr.get_format()
    print "# True barcodes length: " + str(barcode_length)
    print "# Read barcodes length when overhang added: " + str(barcode_length + pre_overhang + post_overhang)



def print_post_stats():
    """
    Prints post stats
    """
    print "# Total reads processed: " + str(stats_total_reads.value())
    cdef int matched_unam = stats_total_reads.value() - stats_unmatched.value() - stats_imperfect_ambiguous_matches.value()
    cdef float tot = float(stats_total_reads.value())
    print "# Reads matched unambiguously: " + str(matched_unam) + " [" + str(100 * matched_unam / tot) + "%]"
    print "#    - Reads matched perfectly: " + str(stats_perfect_matches.value()) + " [" + str(100* stats_perfect_matches.value() / tot) + "%]"
    print "#    - Reads matched imperfectly: " + str(stats_imperfect_unambiguous_matches.value()) + " [" + str(100* stats_imperfect_unambiguous_matches.value() / tot) + "%]"
    print "# Reads matched ambiguously: " + str(stats_imperfect_ambiguous_matches.value()) + " [" + str(100* stats_imperfect_ambiguous_matches.value() / tot) + "%]"
    print "# Reads unmatched: " + str(stats_unmatched.value()) + " [" + str(100* stats_unmatched.value() / tot) + "%]"
    cdef list distr = []
    cdef unsigned int i = 0
    for i in xrange(max_edit_distance+1):
        distr.append(str(stats_edit_distance_counts[i].value()))
    print("# Edit distance counts for [0,...," + str(max_edit_distance) + "]: [" + ", ".join(distr) + "]")



cdef str __match_type_to_str(int match_type):
    """
    Simple converter.
    """
    if match_type == UNMATCHED:
        return "UNMATCHED"
    elif match_type == MATCHED_PERFECTLY:
        return "MATCHED_PERFECTLY"
    elif match_type == MATCHED_UNAMBIGUOUSLY:
        return "MATCHED_UNAMBIGUOUSLY"
    elif match_type == MATCHED_AMBIGUOUSLY:
        return "MATCHED_AMBIGUOUSLY"
    return "KILL"


import os
import argparse
import taggd.io.barcode_utils as bu
import taggd.core.demultiplex_core_functions as core
import taggd.core.demultiplex_search_functions as srch


def main(argv=None):
    """Main application."""

    # Create a parser
    parser = argparse.ArgumentParser(description=__doc__)

    # Needed parameters
    parser.add_argument('barcodes_infile', help="The file with true barcode IDs and other properties.")
    parser.add_argument('reads_infile', help="The FASTQ, FASTA, SAM or BAM file with reads.")
    parser.add_argument('outfile_prefix', help="The output files prefix.")

    # Optional arguments.
    parser.add_argument('--start_position', type=int, help='The start position for barcodes in reads (default: %(default)d)', default=0, metavar="[int]")
    parser.add_argument('--k', type=int, help='The kmer length (default: %(default)d)', default=7, metavar="[int]")
    parser.add_argument('--max_edit_distance', type=int, help='The max edit distance for allowing hits (default: %(default)d)', default=5, metavar="[int]")
    parser.add_argument('--metric', help= "Distance metric: Subglobal, Levenshtein or Hamming (default: %(default)s)", default="Subglobal", metavar="[string]")
    parser.add_argument('--slider_increment', type=int, help="Space between kmer searches, 0 yields kmer length (default: %(default)d)", default=0, metavar="[int]")
    parser.add_argument('--overhang', type=int, help="Additional flanking bases around read barcode to allow for insertions (default: %(default)d)", default=2, metavar="[int]")
    parser.add_argument('--only_output_matched', help="Suppresses writing of output to only file with matched reads.", default=False, action='store_true')
    parser.add_argument('--seed', help="Random number generator seed for shuffling ambiguous hits (default: %(default)s)", default=None, metavar="[string]")
    parser.add_argument('--no_multiprocessing', help="If set, turns off multiprocessing of reads", default=False, action='store_true')
    parser.add_argument('--estimate_min_edit_distance', type=int, help="If set, estimates the min edit distance among true barcodes by comparing the specified number of pairs, 0 means no estimation (default: %(default)d)", default=0, metavar="[int]")

    # Parse
    if argv == None:
        options = parser.parse_args()
    else:
        options = parser.parse_args(argv)

    # Validate all options.
    if not os.path.isfile(options.barcodes_infile) :
        raise ValueError("Invalid true barcodes input file path.")
    if not os.path.isfile(options.reads_infile) :
        raise ValueError("Invalid reads input file path.")
    if not (options.reads_infile.upper().endswith(".FASTQ") or \
                    options.reads_infile.upper().endswith(".FQ") or \
                    options.reads_infile.upper().endswith(".SAM") or \
                    options.reads_infile.upper().endswith(".FASTA") or \
                    options.reads_infile.upper().endswith(".FA") or \
                    options.reads_infile.upper().endswith(".BAM")):
        raise ValueError("Invalid reads input file format: must be FASTQ, FASTA, SAM or BAM format and file end with .fq, fastq, .fa, .fasta, .sam or .bam")
    if options.outfile_prefix is None or options.outfile_prefix == "":
        raise ValueError("Invalid output file prefix.")
    if options.k <= 0:
        raise ValueError("Invalid kmer length. Must be > 0.")
    if options.max_edit_distance < 0:
        raise ValueError("Invalid max edit distance. Must be >= 0.")
    if options.metric not in ("Subglobal", "Levenshtein", "Hamming"):
        raise ValueError("Invalid metric. Must be Subglobal, Levenshtein or Hamming.")
    if options.slider_increment < 0:
        raise ValueError("Invalid slider increment. Must be >= 0.")
    if options.slider_increment == 0:
        options.slider_increment = int(options.k)
    if options.start_position < 0:
        raise ValueError("Invalid start position. Must be >= 0.")
    if options.overhang < 0:
        raise ValueError("Invalid overhang. Must be >= 0.")
    if options.metric == "Hamming" and options.overhang > 0:
        raise ValueError("Invalid overhang. Must be 0 for Hamming metric.")

    # Read barcodes file
    true_barcodes = bu.read_barcode_file(options.barcodes_infile)

    # Check barcodes file.
    if options.estimate_min_edit_distance > 0:
        min_dist = bu.estimate_min_edit_distance(true_barcodes, options.estimate_min_edit_distance)
        if min_dist <= options.max_edit_distance:
            raise ValueError("Invalid max edit distance: exceeds or equal to estimated minimum edit distance among true barcodes.")
        print "# Minimum edit distance between true barcodes was estimated to (may be less) " + str(min_dist)


    # Initialize.
    core.init(true_barcodes, options.reads_infile, os.path.abspath(options.outfile_prefix), options.max_edit_distance, \
              options.start_position, min(options.start_position, options.overhang), options.overhang, options.seed, \
              options.no_multiprocessing, options.only_output_matched)
    srch.init(true_barcodes, options.k, options.max_edit_distance, options.metric, options.slider_increment, \
              min(options.start_position, options.overhang), options.overhang)

    # Demultiplex
    core.print_pre_stats()
    print "# Starting demultiplexing with the following options..." + str(options)
    core.demultiplex()
    print "# ...finished demultiplexing"
    core.print_post_stats()